package kw4h;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

import com.sun.opengl.util.Animator;

/**
 * 
 * @author Kw4h
 *
 */
public class Evemap
{
	// our instance
	private static Evemap instance = null;
	
	// LOGGING
	private Logger logger;
	private Properties log4jSettings = new Properties();
	private FileInputStream log4jStream;
	
	// display objects
	public GLCanvas canvas;
	public JFrame frame;
	private Renderer renderer;
	private MapInputHandler inputHandler = null;
	
	private MapSettings settings = null;
	
	// panel for the left menu
	private JPanel leftPanel;
	
	// subpanel for view settings
	private JPanel viewPanel;
	private JButton flatButton, outlineButton, systemButton, tacticalButton, secstatButton;
	
	// subpanel for planning route
	private JPanel routePanel;
	private JLabel wpLabel, avoidLabel;
	private JComboBox wpBox, avoidBox;
	private JList wpList, avoidList;
	private JButton wpAddButton, wpRemoveButton, avoidAddButton, avoidRemoveButton;
	private ButtonGroup avoidGroup;
	private JRadioButton avoidLowRadio, avoidHighRadio, avoidNoRadio;
	
	// subpanel for system search
	private JPanel searchPanel;
	private JComboBox searchBox;
	private JButton searchButton;
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) 
	{
		String jarUrl = "";
		try {
			jarUrl = new File(new File(System.getProperty("java.class.path")).getParent()).getCanonicalPath();
		} catch (IOException e) {
			jarUrl = "./";
		}

		/*System.out.println(jarUrl);
		System.out.println(System.getProperty("java.library.path"));
		System.out.println(System.getProperty("java.class.path"));*/
		System.setProperty("java.library.path", System.getProperty("java.library.path") + ";" + jarUrl);
		
		System.loadLibrary("gluegen-rt");
		System.loadLibrary("jogl");
		instance = new Evemap();
	}
	
	/**
	 * 
	 * @return
	 */
	public static Evemap getInstance()
	{
		return instance;
	}
	
	/**
	 * 
	 *
	 */
	public Evemap()
	{
		try
		{
			log4jStream = new FileInputStream("log4j.properties");
			log4jSettings.load(log4jStream);
			PropertyConfigurator.configure(log4jSettings);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.exit(0);
		}
		
		logger = Logger.getLogger(this.getClass());
		logger.info("Initiated logger");
		
		settings = MapSettings.getMapSettings();
		inputHandler = new MapInputHandler();
		
		// set system look and feel
		try {
			logger.info("Configuring look and feel");
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (ClassNotFoundException e1) {
			logger.fatal(e1.getMessage());
		} catch (InstantiationException e1) {
			logger.fatal(e1.getMessage());
		} catch (IllegalAccessException e1) {
			logger.fatal(e1.getMessage());
		} catch (UnsupportedLookAndFeelException e1) {
			logger.fatal(e1.getMessage());
		}
		
		// create a frame
		GraphicsEnvironment gEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gDevice = gEnvironment.getDefaultScreenDevice();
		
		this.frame = new JFrame(gDevice.getDefaultConfiguration());
		this.frame.setTitle("EvE Map - v0.5");
		logger.info("Creating frame");
		
		// make a nice window size
		this.frame.setLocation(100, 100);
		this.frame.setSize(1024, 768);
		this.frame.setResizable(true);
		
		this.frame.setLayout(new BorderLayout());
		
		this.renderer = new Renderer(this.frame.getWidth(), this.frame.getHeight());
		this.setupInterface(this.frame.getContentPane());
		this.setupCanvas(this.frame.getContentPane());
		this.frame.setVisible(true);
		logger.info("Showing frame");
		
		// creating an animator
		final Animator animator = new Animator(canvas);
		
		// of course, a window adapter
		logger.info("Adding window listener");
		this.frame.addWindowListener(new WindowAdapter()
		{
			public void windowClosing(WindowEvent e)
			{
				// its been suggested to run this on a different thread
				// otherwise it could be conflicting (animator won't stop before close)
				new Thread(new Runnable()
					{
						public void run()
						{
							logger.info("Closing application");
							animator.stop();
							System.exit(0);
						}
					}				
				).start();
			}
		});
		
		// sleep for a lil bit to draw the interface
		try {
			Thread.sleep(100);
		} catch (InterruptedException e1) {
			logger.fatal(e1.getMessage());
		}
		
		logger.info("Starting Animator");
		animator.setRunAsFastAsPossible(true);
		animator.start();


	}
	
	/**
	 * 
	 * @param container
	 */
	private void setupInterface(Container container)
	{
		// main left panel
		this.leftPanel = new JPanel();
		this.leftPanel.setPreferredSize(new Dimension(180, container.getHeight()));
		container.add(this.leftPanel, BorderLayout.LINE_START);
		
		// sub panel for view settings
		this.viewPanel = new JPanel();
		this.viewPanel.setPreferredSize(new Dimension(170, 160));
		this.viewPanel.setBorder(BorderFactory.createTitledBorder("View"));
		this.leftPanel.add(this.viewPanel);
		
		// add buttons for view settings
		this.flatButton = new JButton("Flatten map");
		this.flatButton.setPreferredSize(new Dimension(110, 20));
		this.flatButton.addActionListener(new BtnListener("flatten"));
		
		this.outlineButton = new JButton("Toggle outlines");
		this.outlineButton.setPreferredSize(new Dimension(110, 20));
		this.outlineButton.addActionListener(new BtnListener("outline"));
		
		this.systemButton = new JButton("System names");
		this.systemButton.setPreferredSize(new Dimension(110, 20));
		this.systemButton.addActionListener(new BtnListener("system"));
		
		this.tacticalButton = new JButton("Tactical view");
		this.tacticalButton.setPreferredSize(new Dimension(110, 20));
		this.tacticalButton.addActionListener(new BtnListener("tactical"));
		
		this.secstatButton = new JButton("Security status");
		this.secstatButton.setPreferredSize(new Dimension(110, 20));
		this.secstatButton.addActionListener(new BtnListener("showsecstat"));
		
		this.viewPanel.add(this.flatButton);
		this.viewPanel.add(this.outlineButton);
		this.viewPanel.add(this.systemButton);
		// this.viewPanel.add(this.tacticalButton);
		this.viewPanel.add(this.secstatButton);
		
		
		// sub panel for route planning
		this.routePanel = new JPanel();
		this.routePanel.setPreferredSize(new Dimension(170, 400));
		this.routePanel.setBorder(BorderFactory.createTitledBorder("Plan route"));
		if (!this.settings.isEveTV())
		{
			this.leftPanel.add(this.routePanel);
		}
		
		// add stuff for route planning
		this.wpLabel = new JLabel("Waypoint:");
		this.wpLabel.setPreferredSize(new Dimension(150, 18));

		this.wpList = new JList();
		this.wpList.setPreferredSize(new Dimension(80, 100));
		this.wpList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		this.wpList.setBorder(BorderFactory.createEtchedBorder(1));
		
		this.wpBox = new JComboBox();
		this.wpBox.setPreferredSize(new Dimension(80, 20));
		this.wpBox.setEditable(true);
		this.wpBox.addActionListener(new BtnListener("combobox", this.wpList, "waypoint"));
		
		this.avoidHighRadio = new JRadioButton("Prefer lowsec");
		this.avoidHighRadio.setPreferredSize(new Dimension(150, 16));
		this.avoidHighRadio.addActionListener(new BtnListener("radiobtn"));
		
		this.avoidLowRadio = new JRadioButton("Prefer highsec");
		this.avoidLowRadio.setPreferredSize(new Dimension(150, 16));
		this.avoidLowRadio.addActionListener(new BtnListener("radiobtn"));
		
		this.avoidNoRadio = new JRadioButton("Don't avoid");
		this.avoidNoRadio.setPreferredSize(new Dimension(150, 16));
		this.avoidNoRadio.setSelected(true);
		this.avoidNoRadio.addActionListener(new BtnListener("radiobtn"));
		
		this.wpAddButton = new JButton("Add");
		this.wpAddButton.setPreferredSize(new Dimension (60, 20));
		this.wpAddButton.addActionListener(new BtnListener("addwp"));
		
		this.wpRemoveButton = new JButton("Del");
		this.wpRemoveButton.setPreferredSize(new Dimension (60, 20));
		this.wpRemoveButton.addActionListener(new BtnListener("remwp"));
		
		this.avoidGroup = new ButtonGroup();
		this.avoidGroup.add(this.avoidNoRadio);
		this.avoidGroup.add(this.avoidLowRadio);
		this.avoidGroup.add(this.avoidHighRadio);
		
		
		// add stuff for route planning for avoiding systems
		this.avoidLabel = new JLabel("Avoid Systems:");
		this.avoidLabel.setPreferredSize(new Dimension(150, 18));
		
		this.avoidList = new JList();
		this.avoidList.setPreferredSize(new Dimension(80, 100));
		this.avoidList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		this.avoidList.setBorder(BorderFactory.createEtchedBorder(1));
		
		this.avoidBox = new JComboBox();
		this.avoidBox.setPreferredSize(new Dimension(80, 20));
		this.avoidBox.setEditable(true);
		this.avoidBox.addActionListener(new BtnListener("combobox", this.avoidList, "avoid"));
		
		this.avoidAddButton = new JButton("Add");
		this.avoidAddButton.setPreferredSize(new Dimension (60, 20));
		this.avoidAddButton.addActionListener(new BtnListener("addavoid"));
		
		this.avoidRemoveButton = new JButton("Del");
		this.avoidRemoveButton.setPreferredSize(new Dimension (60, 20));
		this.avoidRemoveButton.addActionListener(new BtnListener("remavoid"));
		
		
		this.routePanel.add(this.wpLabel);
		this.routePanel.add(this.wpBox);
		this.routePanel.add(this.wpAddButton);
		this.routePanel.add(this.wpList);
		this.routePanel.add(this.wpRemoveButton);
		this.routePanel.add(this.avoidNoRadio);
		this.routePanel.add(this.avoidLowRadio);
		this.routePanel.add(this.avoidHighRadio);
		
		this.routePanel.add(this.avoidLabel);
		this.routePanel.add(this.avoidBox);
		this.routePanel.add(this.avoidAddButton);
		this.routePanel.add(this.avoidList);
		this.routePanel.add(this.avoidRemoveButton);
		
		
		// subpanel for system search
		this.searchPanel = new JPanel();
		this.searchPanel.setPreferredSize(new Dimension(170, 60));
		this.searchPanel.setBorder(BorderFactory.createTitledBorder("Find a system"));
		this.leftPanel.add(this.searchPanel);
		
		this.searchBox = new JComboBox();
		this.searchBox.setPreferredSize(new Dimension(80, 20));
		this.searchBox.setEditable(true);
		this.searchBox.addActionListener(new BtnListener("combobox"));
		
		this.searchButton = new JButton("Find");
		this.searchButton.setPreferredSize(new Dimension (60, 20));
		this.searchButton.addActionListener(new BtnListener("searchsystem"));
		
		this.searchPanel.add(this.searchBox);
		this.searchPanel.add(this.searchButton);
	}
	
	/**
	 * 
	 * @param container
	 */
	private void setupCanvas(Container container)
	{
		// Create opengl canvas, and add it to the frame
		GLCapabilities capabilities = new GLCapabilities();
		capabilities.setHardwareAccelerated(true);
		capabilities.setDoubleBuffered(true);
		
		this.canvas = new GLCanvas(capabilities);
		container.add(canvas);
		this.canvas.addGLEventListener(this.renderer);
		this.canvas.addMouseListener(this.inputHandler);
		this.canvas.addMouseMotionListener(this.inputHandler);
		this.canvas.addKeyListener(this.inputHandler);
		this.canvas.addMouseWheelListener(this.inputHandler);
	}
	
	/**
	 * 
	 *
	 */
	public void toggleFullscreen()
	{
		logger.debug("Current fullscreenmode: " + this.settings.getIsFullscreen());
		
		// windowed -> fullscreen
		if (this.settings.getIsFullscreen() == false)
		{
			logger.debug("Setting fullscreenmode to: " + true);
			this.settings.setIsFullscreen(true);
			
			// remove decoration
			this.frame.dispose();
			this.frame.setUndecorated(true);
			this.frame.setVisible(true);
			
			// get graphics device
			GraphicsDevice gDevice = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
			gDevice.setFullScreenWindow(this.frame);
			
		}
		// fullscreen -> windowed
		else
		{
			logger.debug("Setting fullscreenmode to: " + false);
			this.settings.setIsFullscreen(false);
			
			// remove decoration
			this.frame.dispose();
			this.frame.setUndecorated(false);
			
			// get graphics device
			GraphicsDevice gDevice = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
			gDevice.setFullScreenWindow(null);
			
			this.frame.setVisible(true);
		}
		
	}
	
	/**
	 * 
	 *
	 */
	public void toggleLeftPanel()
	{
		this.leftPanel.setVisible(!this.leftPanel.isVisible());
	}
	
	/**
	 * 
	 * @author Kw4h
	 *
	 */
	private class BtnListener implements ActionListener
	{
		private String action;
		private JList listBox = null;
		private String type;
		
		/**
		 * 
		 * @param action
		 */
		public BtnListener(String action)
		{
			this.action = action;
		}
		
		/**
		 * 
		 * @param action
		 * @param list
		 * @param type
		 */
		public BtnListener(String action, JList list, String type)
		{
			this.action = action;
			this.listBox = list;
			this.type = type;
		}
		
		/**
		 * 
		 */
		public void actionPerformed(ActionEvent e) 
		{
			Universe universe = Universe.getUniverse();
			
			// do stuff based on action
			// flatten map
			if (this.action.equals("flatten"))
			{
				settings.toggleFlat();
			}
			// outline
			else if (this.action.equals("outline"))
			{
				settings.toggleOutline();
			}
			// draw system names
			else if (this.action.equals("system"))
			{
				settings.toggleSystem();
			}
			// toggle tactical
			else if (this.action.equals("tactical"))
			{
				settings.toggleTactical();
			}
			// pressing enter on combobox
			else if (this.action.equals("combobox"))
			{				
				if (e.getActionCommand().equals("comboBoxEdited"))
				{
					// get what is put in
					JComboBox source = (JComboBox) e.getSource();
					String input = (String) source.getSelectedItem();
					source.removeAllItems();
					
					// only two letters or more
					if (input.length() >= 2)
					{
						Vector found = DBReader.getDBReader().findSolarSystem(input);
						
						if (found.size() > 1)
						{
							source.addItem(input);
						}
						
						for (Iterator i = found.iterator(); i.hasNext();)
						{
							source.addItem(i.next());
						}
						
						// show popup
						if (found.size() > 1)
						{
							source.showPopup();
						}
						// if there's only one result, and our listbox is set
						else if ((found.size() == 1) && this.listBox != null)
						{
							Vector syslist;
							if (this.type.equals("waypoint"))
							{
								syslist = universe.getTravelSystems();
							}
							else
							{
								syslist = universe.getAvoidSystems();
							}
							syslist.add(found.firstElement());
							
							// update the list
							this.listBox.setListData(syslist);
							
							int avoidance = 0;
							if (avoidLowRadio.isSelected())
							{
								avoidance = 1;
							}
							else if (avoidHighRadio.isSelected())
							{
								avoidance = 2;
							}
							
							// start calculation
							universe.calcWaypoints(avoidance);
						}
					}
					// if there's only one or less letters
					else
					{
						source.addItem(input);
					}
				}
			}
			// hit the search system button
			else if (this.action.equals("searchsystem"))
			{
				String sys = (String) searchBox.getSelectedItem();
				Universe.getUniverse().setSearchedSystem(sys);
				
				searchBox.removeAllItems();
			}
			// hit the add wp button
			else if (this.action.equals("addwp"))
			{
				// get the system name and add to the list
				String sys = (String) wpBox.getSelectedItem();
				if (sys != null)
				{
					universe.addTravelSystem(sys);
					
					// update the list
					wpList.setListData(universe.getTravelSystems());
					
					int avoidance = 0;
					if (avoidLowRadio.isSelected())
					{
						avoidance = 1;
					}
					else if (avoidHighRadio.isSelected())
					{
						avoidance = 2;
					}
					
					// start calculation
					universe.calcWaypoints(avoidance);
				}
				
				// clear input box
				wpBox.removeAllItems();
			}
			else if (this.action.equals("remwp"))
			{
				// remove item from list, and update the view
				universe.removeTravelSystem(wpList.getSelectedValue().toString());
				wpList.setListData(universe.getTravelSystems());
				
				// recalculate
				int avoidance = 0;
				if (avoidLowRadio.isSelected())
				{
					avoidance = 1;
				}
				else if (avoidHighRadio.isSelected())
				{
					avoidance = 2;
				}
				
				universe.calcWaypoints(avoidance);
			}
			// hit the add avoid button
			else if (this.action.equals("addavoid"))
			{
				// get the system name and add to the list
				String sys = (String) avoidBox.getSelectedItem();
				if (sys != null)
				{
					universe.addAvoidSystem(sys);
					
					// update the list
					avoidList.setListData(universe.getAvoidSystems());
					
					int avoidance = 0;
					if (avoidLowRadio.isSelected())
					{
						avoidance = 1;
					}
					else if (avoidHighRadio.isSelected())
					{
						avoidance = 2;
					}
					
					// start calculation
					universe.calcWaypoints(avoidance);
				}
				
				// clear input box
				avoidBox.removeAllItems();
			}
			else if (this.action.equals("remavoid"))
			{
				// remove item from list, and update the view
				universe.removeAvoidSystem(avoidList.getSelectedValue().toString());
				avoidList.setListData(universe.getAvoidSystems());
				
				// recalculate
				int avoidance = 0;
				if (avoidLowRadio.isSelected())
				{
					avoidance = 1;
				}
				else if (avoidHighRadio.isSelected())
				{
					avoidance = 2;
				}
				
				universe.calcWaypoints(avoidance);
			}
			else if (this.action.equals("showsecstat"))
			{
				settings.toggleSecStat();
			}
			else if (this.action.equals("radiobtn"))
			{
				// recalculate
				int avoidance = 0;
				if (avoidLowRadio.isSelected())
				{
					avoidance = 1;
				}
				else if (avoidHighRadio.isSelected())
				{
					avoidance = 2;
				}
				
				universe.calcWaypoints(avoidance);
			}
		}
	}
}
