package kw4h;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import org.apache.log4j.Logger;

public class MapInputHandler implements MouseListener, MouseMotionListener, KeyListener, MouseWheelListener
{
	private Logger logger = Logger.getLogger(this.getClass());
	
	// settings
	private MapSettings settings = MapSettings.getMapSettings();
	
	public void mouseClicked(MouseEvent arg0) {
	}

	public void mouseEntered(MouseEvent arg0) {
	}

	public void mouseExited(MouseEvent arg0) {
	}

	public void mousePressed(MouseEvent e) 
	{
		// increase buttons pressed
		this.settings.btnsPressed += 1;
		
		// only left OR right button
		if (e.getButton() == 1 || e.getButton() == 3 && this.settings.btnsPressed == 1)
		{
			// if the startx and y are -1, it means its not set
			if (this.settings.mStartX == -1 && this.settings.mStartY == -1)
			{
				this.settings.mStartX = e.getX();
				this.settings.mStartY = e.getY();
			}
		}
		// 2 or more buttons pressed
		else if (this.settings.btnsPressed >= 2)
		{
			// set the startx and y
			this.settings.mStartX = e.getX();
			this.settings.mStartY = e.getY();
		}
	}

	public void mouseReleased(MouseEvent e) 
	{
		// reset startx and y to -1
		this.settings.mStartX = -1;
		this.settings.mStartY = -1;
		
		// decrease buttons pressed
		this.settings.btnsPressed -= 1;
	}

	public void mouseDragged(MouseEvent e) 
	{
		// if the startX and Y are set, mouse is pressed
		if (this.settings.mStartX != -1 && this.settings.mStartY != -1 && this.settings.btnsPressed == 1)
		{
			int newx, newy;
			newx = e.getX();
			newy = e.getY();
			
			// calculate how much to move
			this.settings.mMoveX = (newx - this.settings.mStartX) * -1;
			this.settings.mMoveZ = newy - this.settings.mStartY;
			
			// set the new mStartX and mStartY
			this.settings.mStartX = newx;
			this.settings.mStartY = newy;
		}
		// modify the scaling - aka zooming in
		else if (this.settings.mStartX != -1 && this.settings.mStartY != -1 && this.settings.btnsPressed >= 2)
		{
			// get the x and y
			int newx, newy;
			newx = e.getX();
			newy = e.getY();
			
			// make new scale
			this.settings.mMoveY = (newy - this.settings.mStartY) * -1;
			
			// set the new mStartX and mStartY
			this.settings.mStartX = newx;
			this.settings.mStartY = newy;
		}
	}

	public void mouseMoved(MouseEvent e) 
	{
		if (e.getButton() == 0)
		{
			this.settings.btnsPressed = 0;
		}
	}

	public void keyPressed(KeyEvent e) 
	{
		// button was pressed (duh)
		logger.debug("Keycode: " + e.getKeyCode());
		
		switch(e.getKeyCode())
		{
			// the f button
			// toggle flat or depth mode
			case 70:
				this.settings.toggleFlat();
				break;
			
			// the o button
			// toggle draw outlines
			case 79:
				this.settings.toggleOutline();
				break;
				
			// the s button
			case 83:
				this.settings.toggleSecStat();
				break;
				
			// the n button
			case 78:
				this.settings.toggleSystem();
				break;
				
			// zooming in
			// the A button
			case 65:
				this.settings.zoomIn();
				break;
				
			// the + button
			case 61:
				this.settings.zoomIn();
				break;
				
			// zooming out
			// the Z button
			case 90:
				this.settings.zoomOut();
				break;
				
			// the - button
			case 45:
				this.settings.zoomOut();
				break;
				
			// enable grid
			// the g button
			case 71:
				this.settings.toggleGrid();
				break;
			
			// fullscreen - k button	
			case 75:
				Evemap.getInstance().toggleFullscreen();
				break;
				
			// leftpanel - l button	
			case 76:
				Evemap.getInstance().toggleLeftPanel();
				break;
		}
	}

	public void keyReleased(KeyEvent e) {
	}

	public void keyTyped(KeyEvent e) {
	}

	public void mouseWheelMoved(MouseWheelEvent e) 
	{
		this.settings.mMoveY = e.getWheelRotation() * -30;
	}
}
