package kw4h;
// TODO: constellation names on distance
// TODO: calculate cyno jumps
 
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;

import org.apache.log4j.Logger;

import com.sun.opengl.util.GLUT;

/**
 * 
 * @author Kw4h
 *
 */
public class Renderer implements GLEventListener
{	
	private GLU glu;
	private GLUT glut;
	
	private boolean isReady = false;
	private MapSettings settings = MapSettings.getMapSettings();
	
	private Logger logger = Logger.getLogger(this.getClass());
	
	/**
	 * Constructor
	 *
	 */
	public Renderer(int width, int height)
	{
		logger.info("Loaded class Renderer");
		
		this.settings.offsetX = (double) (width);
		this.settings.offsetZ = (double) (height);
		
		this.glu = new GLU();
		this.glut = new GLUT();
	}
	
	/**
	 * Initialising of the opengl world
	 */
	public void init(GLAutoDrawable drawable)
	{
		
	}
	
	/**
	 * 
	 */
	public void display(GLAutoDrawable drawable) 
	{
		if (this.isReady == false)
		{
			GL gl = drawable.getGL();			
			this.doInit(gl);
		}
		
		if (this.isReady == true)
		{
			// get the GL object
			GL gl = drawable.getGL();
			
			// clear the buffer
			gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
			
			// setup the view
			if (this.settings.mMoveX != 0 || this.settings.mMoveY != 0 || this.settings.mMoveZ != 0)
			{
				this.setupView(gl);
			}

			// set the modelview
			gl.glMatrixMode(GL.GL_MODELVIEW);
			gl.glLoadIdentity();
			
			// set the heightScale
			this.setHeightScale();
			
			this.drawGrid(gl);
			this.drawSystemJumps(gl);
			this.drawSolarSystems(gl);
			this.drawRegions(gl);
			
			if (!this.settings.isEveTV())
			{
				this.drawTravelPath(gl);
				
				// 2d stuff
				this.draw2d(gl);
			}			
			
			gl.glFlush();
		}
	}

	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
	}

	/**
	 * Resizing of viewport
	 */
	public void reshape(GLAutoDrawable drawable, int x, int y, int width,int height) 
	{
		// ge the GL object
		GL gl = drawable.getGL();
		
		// set up the projection (lens for camera)
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		this.glu.gluPerspective(60.0f, (double) width/height, 5.0, 800000.0);
		gl.glTranslated(this.settings.viewX, this.settings.viewY, this.settings.viewZ);
		gl.glViewport(0, 0, width, height);
		
		// offsets
		this.settings.offsetX = (double) (width);
		this.settings.offsetZ = (double) (height);
		
		this.setupView(gl);
		
		doGlInit(gl);
	}

	/* ==============================================*
	 * Custom methods start here, like drawing a line*
	 * ==============================================*/
	
	/**
	 * 
	 */
	private void doInit(GL gl)
	{	
		logger.info("Running initialisation process");
		DBReader reader = DBReader.getDBReader();
		//reader.doImport();
		Universe universe = Universe.getUniverse();
		
		universe.setSolarSystems(reader.getSolarSystems());
		universe.setSystemJumps(reader.getSystemJumps());
		universe.setRegions(reader.getRegions());
		
		// default mouse vars
		this.settings.mStartX = -1;
		this.settings.mStartY = -1;
		this.settings.mMoveX = 1;
		this.settings.mMoveY = 0;
		this.settings.mMoveZ = 0;
		
		// default view location
		this.settings.viewX = 0;
		this.settings.viewY = -500;
		this.settings.viewZ = 0;
		
		doGlInit(gl);
		
		this.glu.gluPerspective(60.0f, this.settings.offsetX / this.settings.offsetZ, 1.0f, 500.0f);
		
		this.isReady = true;
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void doGlInit(GL gl)
	{
		// set bg to black - and rendering to flat
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		gl.glClearDepth(1.0f);
		
		// sexy lines - enabling smooth lines and funky blending at the nicest quality
		gl.glEnable(GL.GL_LINE_SMOOTH);
		gl.glEnable(GL.GL_BLEND);
		gl.glDisable(GL.GL_DITHER);
		gl.glBlendFunc( GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glShadeModel(GL.GL_SMOOTH);
		gl.glHint(GL.GL_LINE_SMOOTH_HINT, GL.GL_NICEST);
		
		gl.glEnable(GL.GL_DEPTH_TEST);
		// gl.glDepthFunc(GL.GL_LEQUAL);
		gl.glDepthFunc(GL.GL_ALWAYS);
	}
	
	/**
	 * 
	 */
	private void setupView(GL gl)
	{
		
		// load projection matrix
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		this.glu.gluPerspective(45.0f, this.settings.offsetX / this.settings.offsetZ, 5.0, 800000.0);
		gl.glViewport(0, 0, (int) this.settings.offsetX, (int) this.settings.offsetZ);
		
		// set new x, y and z for the cam
		double accel = ((double) this.settings.viewY / 825);
		if (accel >= -0.2)
		{
			accel = -0.2;
		}
		
		this.settings.viewX += (this.settings.mMoveX * -accel);
		this.settings.viewY += (this.settings.mMoveY * accel);
		this.settings.viewZ += (this.settings.mMoveZ * -accel);
		
		// we're not going any lower than -15
		//if (this.settings.viewZ > 0)
		//{
		//	this.settings.viewZ = 0;
		//}
		
		// set the move to 0
		this.settings.mMoveX = 0;
		this.settings.mMoveY = 0;
		this.settings.mMoveZ = 0;
		
		// transform
		// gl.glTranslated(this.settings.viewX, this.settings.viewY, this.settings.viewZ);
		
		// glu.gluLookAt(0, -500, 250, 0, 0, 250, 0, 0, 1);
		glu.gluLookAt(settings.viewX, settings.viewY, settings.viewZ, settings.viewX, settings.viewY + 1.0f, settings.viewZ, 0, 0, 1);
		//gl.glRotatef(-90, 1, 0, 0);
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void draw2d(GL gl)
	{
		// set up the ortho view for 2d drawing
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPushMatrix();
		gl.glLoadIdentity();
		gl.glOrtho(0, this.settings.offsetX, this.settings.offsetZ, 0, -1, 1);
		gl.glMatrixMode(GL.GL_MODELVIEW);
		
		// do the actual drawing here
		// draw jumps, if any
		this.drawJumpsMenu(gl);
		
		// draw shortcut menu
		this.drawShorts(gl);
		
		// draw copyright
		this.drawCopyright(gl);
		
		// set the view back
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPopMatrix();
		gl.glMatrixMode(GL.GL_MODELVIEW);
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawJumpsMenu(GL gl)
	{
		// only draw if there's any jumps
		Universe universe = Universe.getUniverse();
		if (universe.getTravelJumps().size() > 0)
		{
			// set the color
			gl.glColor3f(0.9f, 0.9f, 0.9f);
			
			// begin drawing the outlines
			gl.glPointSize(1.0f);
			gl.glBegin(GL.GL_LINES);
			
			// horizontal line
			gl.glVertex2f(10, 10);
			gl.glVertex2f(130, 10);
			gl.glEnd();
			
			// loop through all jumps
			int num = 1;
			int wpnum = 1;
			int jumpNum = 0;
			for (Iterator y = universe.getTravelJumps().iterator(); y.hasNext();)
			{
				Vector v = (Vector) y.next();
				jumpNum += (v.size() - 1);
				for (Iterator i = v.iterator(); i.hasNext();)
				{
					SolarSystem sys = (SolarSystem) i.next();
					
					if (sys != null)
					{
						// get security status
						double security = sys.security;
						if (security < 0) security = 0.0;
						
						String sec = String.valueOf(security).substring(0, 3);
						
						// draw the solarsystem name - and sec status
						gl.glRasterPos2d(15, (16 * num) + 10);
						this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, sys.name);
						
						gl.glRasterPos2d(100, (16 * num) + 10);
						this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(" + sec + ")");
						
						// if this is a waypoint
						if (universe.getTravelSystems().contains(sys.name))
						{
							gl.glRasterPos2d(130, (16 * num) + 10);
							this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "Waypoint " + String.valueOf(wpnum));
							wpnum++;
						}
						
						num ++;
					}
				}
			}
			
			// display the total jumps
			gl.glRasterPos2d(190, (16 * 1) + 10);
			this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "Jumps: " + String.valueOf(jumpNum));
			
			// vertical line
			gl.glBegin(GL.GL_LINES);
			gl.glVertex2f(10, 10);
			gl.glVertex2f(10, (num * 16) + 10);
			gl.glEnd();
		}
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawShorts(GL gl)
	{
		// draw menu border
		gl.glColor3f(0.9f, 0.9f, 0.9f);
		gl.glBegin(GL.GL_LINES);
		
		// horizontal line
		gl.glVertex2f((float)this.settings.offsetX - 10, 10);
		gl.glVertex2f((float)this.settings.offsetX - 170, 10);
		
		gl.glEnd();
		
		// start drawing text
		int i = 1;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(f) Flatten map: " + String.valueOf(this.settings.drawFlat));
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(o) Draw outlines: " + String.valueOf(this.settings.drawOutlines));
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(s) Security status: " + String.valueOf(this.settings.showSecStat));
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(n) Solarsystem names: " + String.valueOf(this.settings.drawSystemNames));
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(g) Show grid: " + String.valueOf(this.settings.drawGrid));
		i++;
		
		i++;
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(k) Toggle fullscreen");
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(l) Toggle left panel");
		i++;
		
		i++;
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(z or -) Zoom out");
		i++;
		
		gl.glRasterPos2d(this.settings.offsetX - 170, (16 * i) + 10);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "(a or +) Zoom in");
		i++;
		
		// vertical line
		gl.glBegin(GL.GL_LINES);
		gl.glVertex2f((float)this.settings.offsetX - 10, 10);
		gl.glVertex2f((float)this.settings.offsetX - 10, (16 * (i) + 10));
		gl.glEnd();
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawCopyright(GL gl)
	{
		gl.glColor4f(1, 1, 1, 0.2f);
		gl.glRasterPos2d(this.settings.offsetX - 70, this.settings.offsetZ - 5);
		this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "2007 Kw4h");
	}
	
	/**
	 * 
	 */
	private void drawLine(GL gl, double[] startPos, double[] endPos, float R, float G, float B)
	{
		// set the color
		gl.glColor3f(R, G, B);
		
		// begin the draw session
		gl.glBegin(GL.GL_LINES);
		
		gl.glVertex3d(startPos[0], startPos[1] * settings.heightScale, startPos[2]);
		gl.glVertex3d(endPos[0], endPos[1] * settings.heightScale, endPos[2]);
		
		// end the draw session
		gl.glEnd();
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawGrid(GL gl)
	{
		if (this.settings.drawGrid)
		{
			gl.glColor4f(0.2f, 0.4f, 1.0f, 0.25f);
			
			// draw vertical lines
			gl.glBegin(GL.GL_LINES);
			for (int i = 0; i < 81; i ++)
			{
				double x = (-1200 + (25 * i));
				
				// draw the two points
				gl.glVertex3d(x, 0, -1000);
				gl.glVertex3d(x, 0, 1100);
				
			}
			
			// draw horizontal lines
			for (int z = 0; z < 85; z ++)
			{
				double y = (-1000 + (25 * z));
				
				// draw the two points
				gl.glVertex3d(-1200, 0, y);
				gl.glVertex3d(800, 0, y);
			}
			
			gl.glEnd();
		}
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawSolarSystems(GL gl)
	{	
		// FIXME: needs optimizing
		// system dots always on top
		gl.glDisable(GL.GL_DEPTH_TEST);
		boolean colorChanged = true;
		
		// loop through solarsystems
		Enumeration num;
		num = Universe.getUniverse().getSolarSystems().elements();
		
		// start drawing
		gl.glColor3f(0.4f, 0.4f, 0.6f);
		gl.glPointSize(2.0f);
		gl.glBegin(GL.GL_POINTS);
		
		while(num.hasMoreElements())
		{
			// get the solarsystem element
			SolarSystem sys = (SolarSystem) num.nextElement();
			
			// draw the found system
			if (sys.name.equalsIgnoreCase(Universe.getUniverse().getSearchedSystem()))
			{
				this.drawFoundSystem(gl, sys);
			}
			
			// default color
			if (colorChanged == true)
			{
				gl.glEnd();
				gl.glColor3f(0.4f, 0.4f, 0.6f);
				gl.glPointSize(2.0f);
				gl.glBegin(GL.GL_POINTS);
				colorChanged = false;
			}
			
			// get security status
			if (this.settings.showSecStat)
			{
				gl.glEnd();
				gl.glPointSize(3f);
				gl.glColor3dv(sys.color, 0);
				gl.glBegin(GL.GL_POINTS);
				colorChanged = true;
			}
			
			// add a single point - normal points upwards
			gl.glVertex3d(sys.position[0], sys.position[1] * settings.heightScale, sys.position[2]);
			
			// check if the distance is lower than sysrange
			// FIXME
			if (this.settings.drawSystemNames)
			{
				double[] viewPos = new double[3];
				viewPos[0] = this.settings.viewX;
				viewPos[1] = this.settings.viewY;
				viewPos[2] = this.settings.viewZ;
				
				if (KwCalc.calcDistance(sys.position, viewPos) < this.settings.SYSRANGE)
				{
					// draw the text
					gl.glEnd();
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glRasterPos3d(sys.position[0], sys.position[1], sys.position[2]);
					this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, sys.name);
					colorChanged = true;
					gl.glBegin(GL.GL_POINTS);
				}
			}
			
			// if we are avoiding this system
			if (Universe.getUniverse().getAvoidSystems().contains(sys.name))
			{
				// draw the text
				gl.glEnd();
				gl.glColor3f(0.5f, 0.5f, 0.5f);
				gl.glRasterPos3dv(sys.position, 0);
				this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, sys.name);
				
				double[] textPosition = sys.position.clone();
				textPosition[2] -= 5f;
				
				gl.glColor3f(1f, 0f, 0f);
				gl.glRasterPos3dv(textPosition, 0);
				this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, "Avoiding");
				colorChanged = true;
				gl.glBegin(GL.GL_POINTS);
			}
		}
		
		// end drawing
		gl.glEnd();
		
		gl.glEnable(GL.GL_DEPTH_TEST);
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawSystemJumps(GL gl)
	{
		// systemJumps: SolarSystem,vector
		// vector has all destinations
		// loop through solarsystems
		Enumeration num;
		num = Universe.getUniverse().getSystemJumps().keys();

		// loop through all the from's
		while(num.hasMoreElements())
		{
			SolarSystem from = (SolarSystem) num.nextElement();
			Vector toSystems;
			toSystems = (Vector) Universe.getUniverse().getSystemJumps().get(from);
			
			// loop to all destinations
			Iterator it = toSystems.iterator();
			while(it.hasNext())
			{
				SolarSystem to = (SolarSystem) it.next();
				
				// linecolor
				float r, g, b;
				
				// region jump
				if (from.regionID != to.regionID)
				{
					r = 0.5f;
					g = 0.0f;
					b = 0.5f;
				}
				// constellationjump
				else if (from.constellationID != to.constellationID)
				{
					r = 0.5f;
					g = 0.0f;
					b = 0.0f;
				}
				// normal jump
				else
				{
					r = 0.0f;
					g = 0.0f;
					b = 0.5f;
				}
				
				// draw each line
				this.drawLine(gl, from.position, to.position, r, g, b);

				
				gl.glEnable(GL.GL_DEPTH_TEST);
			}
		}
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawTravelPath(GL gl)
	{
		// travel jump
		float r, g, b;
		r = 1.0f;
		g = 0.5f;
		b = 0.0f;
		
		// for each vector in this vector
		for (Iterator i = Universe.getUniverse().getTravelJumps().iterator(); i.hasNext();)
		{
			// each vector holds solarsystems
			Vector vec = (Vector) i.next();
			
			SolarSystem prev = null;
			SolarSystem next = null;
			for (Iterator y = vec.iterator(); y.hasNext();)
			{
				if (prev == null)
				{
					prev = (SolarSystem) y.next();
					// draw the text
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glRasterPos3d(prev.position[0], prev.position[1] * settings.heightScale, prev.position[2]);
					this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, prev.name);
				}
				
				// get next system
				next = (SolarSystem) y.next();
				
				// draw line
				if (prev != null && next != null)
				{
					this.drawLine(gl, prev.position, next.position, r, g, b);
					
					// draw the text
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glRasterPos3d(next.position[0], next.position[1] * settings.heightScale, next.position[2]);
					this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_10, next.name);
				}
				
				// next becomes prev
				prev = next;
			}
		}
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawFoundSystem(GL gl, SolarSystem sys)
	{
		float r, g, b;
		r = 1.0f;
		g = 0.5f;
		b = 0.0f;
		
		double[] lineTop = sys.position.clone();
		lineTop[1] -= 100000;
		double[] lineBottom = sys.position.clone();
		lineBottom[1] += 100000;
		
		double[] lineLeft = sys.position.clone();
		lineLeft[2] -= 1000000;
		double[] lineRight = sys.position.clone();
		lineRight[2] += 1000000;
		
		this.drawLine(gl, lineTop, lineBottom, r, g, b);
		this.drawLine(gl, lineLeft, lineRight, r, g, b);
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawRegions(GL gl)
	{
		Enumeration num = Universe.getUniverse().getRegions().elements();
		
		// loop through all regions
		while(num.hasMoreElements())
		{
			// get the region
			Region region = (Region) num.nextElement();
			
			if (region != null)
			{	
				// only draw outlines when enabled
				if (this.settings.drawOutlines == true)
				{
					// region has to be within a certain angle
					// if (KwCalc.calcDistance(-region.x, -region.z, (region.y * this.settings.heightScale), this.settings.viewX, this.settings.viewY, -this.settings.viewZ) < this.settings.REGIONRANGE)
					double[] viewPos = new double[3];
					viewPos[0] = this.settings.viewX;
					viewPos[1] = this.settings.viewY;
					viewPos[2] = this.settings.viewZ;
					
					if (KwCalc.calcDistance(region.position, viewPos) < this.settings.REGIONRANGE)
					{
						double angle = KwCalc.calcViewAngle(this.settings.viewX, this.settings.viewY, -this.settings.viewZ, -region.x, -region.z, region.y);
						if (angle < 25 && angle > -25)
						{
							this.drawConvex(gl, region);
							
							// only draw region name when outlined
							// draw the text - make sure its always on top
							gl.glDisable(GL.GL_DEPTH_TEST);
							gl.glColor3f(0.5f, 0.5f, 0.5f);
							gl.glRasterPos3d((region.x), (region.z), (region.y * this.settings.heightScale));
							this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_12, region.name);
							gl.glEnable(GL.GL_DEPTH_TEST);
						}
					}
				}
				else
				{
					// draw the text - make sure its always on top
					gl.glDisable(GL.GL_DEPTH_TEST);
					gl.glColor3f(0.5f, 0.5f, 0.5f);
					gl.glRasterPos3d(region.position[0], region.position[1] * settings.heightScale, region.position[2]);
					this.glut.glutBitmapString(GLUT.BITMAP_HELVETICA_12, region.name);
					gl.glEnable(GL.GL_DEPTH_TEST);
				}
			}
		}
	}
	
	/**
	 * 
	 *
	 */
	private void setHeightScale()
	{
		// TODO: make it timebound, not fps bound
		// in 3d mode, going to flat
		if (this.settings.drawFlat == true && this.settings.heightScale > 0)
		{
			this.settings.heightScale -= 0.01;
			if (this.settings.heightScale < 0)
			{
				this.settings.heightScale = 0;
			}
		}
		// in flat mode, going to 3d mode
		else if (this.settings.drawFlat == false && this.settings.heightScale < 1)
		{
			this.settings.heightScale += 0.01;
			if (this.settings.heightScale > 1)
			{
				this.settings.heightScale = 1;
			}
		}
	}
	
	/**
	 * 
	 * @param gl
	 */
	private void drawConvex(GL gl, Region curRegion)
	{
		
	}
	
}
