/**
 * 
 */
package kw4h;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * @author Kw4h
 *
 */
public class Universe 
{
	private static Universe instance = null;
	
	private Logger logger = Logger.getLogger(this.getClass());
	
	// global storage
	private Hashtable solarSystems = new Hashtable();
	private Hashtable systemJumps = new Hashtable();
	private Hashtable regions = new Hashtable();
	
	private int avoidance = 0;
	
	private String searchedSystem = null;
	
	// for traveling
	private Vector travelJumps = new Vector();
	private Vector travelSystems = new Vector();
	private Vector avoidSystems = new Vector();
	
	/**
	 * 
	 * @return
	 */
	public static Universe getUniverse()
	{
		if (instance == null)
		{
			instance = new Universe();
		}
		
		return instance;
	}
	
	/**
	 * 
	 * @param sys
	 */
	public void setSearchedSystem(String sys)
	{
		logger.info("Creating isntance");
		if (DBReader.getDBReader().systemExists(sys))
		{
			this.searchedSystem = sys;
		}
		else
		{
			this.searchedSystem = null;
		}
	}
	
	public String getSearchedSystem()
	{
		return this.searchedSystem;
	}
	
	/**
	 * 
	 * @param systems
	 */
	public void setSolarSystems(Hashtable systems)
	{
		this.solarSystems = (Hashtable) systems.clone();
	}
	
	/**
	 * 
	 * @return
	 */
	public Hashtable getSolarSystems()
	{
		return this.solarSystems;
	}
	
	/**
	 * 
	 * @param jumps
	 */
	public void setSystemJumps(Hashtable jumps)
	{
		this.systemJumps = (Hashtable) jumps.clone();
	}
	
	/**
	 * 
	 * @return
	 */
	public Hashtable getSystemJumps()
	{
		return this.systemJumps;
	}
	
	/**
	 * 
	 * @param travelJumps
	 */
	public void setTravelJumps(Vector travelJumps)
	{
		this.travelJumps = travelJumps;
	}
	
	/**
	 * 
	 * @return
	 */
	public Vector getTravelJumps()
	{
		return this.travelJumps;
	}
	
	/**
	 * 
	 * @param travelSystems
	 */
	public void setTravelSystems(Vector travelSystems)
	{
		this.travelSystems = travelSystems;
	}
	
	public Vector getTravelSystems()
	{
		return this.travelSystems;
	}
	
	public void addTravelSystem(String systemName)
	{
		if (DBReader.getDBReader().systemExists(systemName) && !this.travelSystems.contains(systemName))
		{
			this.travelSystems.add(systemName);
		}
	}
	
	public void removeTravelSystem(String systemName)
	{
		this.travelSystems.remove(systemName);
	}
	
	
	/**
	 * 
	 * @param avoidSystems
	 */
	public void setAvoidSystems(Vector avoidSystems)
	{
		this.avoidSystems = avoidSystems;
	}

	public Vector getAvoidSystems()
	{
		return this.avoidSystems;
	}
	
	public void addAvoidSystem(String systemName)
	{
		if (DBReader.getDBReader().systemExists(systemName) && !this.avoidSystems.contains(systemName))
		{
			this.avoidSystems.add(systemName);
		}
	}
	
	public void removeAvoidSystem(String systemName)
	{
		this.avoidSystems.remove(systemName);
	}
	
	/**
	 * 
	 * @param avoidance
	 */
	public void calcWaypoints(int avoidance)
	{
		logger.info("Starting waypoint calculation");
		
		// this will take all waypoints, and calculate the path between them
		this.avoidance = avoidance;
		this.travelJumps.clear();
		
		// only if there's enough systems
		if (this.travelSystems.size() >= 2)
		{
			String prev = null;
			String next = null;
			for (Iterator i = this.travelSystems.iterator(); i.hasNext();)
			{
				if (prev == null)
				{
					prev = (String) i.next();
				}
				
				// get the next system
				next = (String) i.next();
				
				// Calculate the waypoint
				Vector path = this.getTravelPath(next, prev);
				if (path != null)
				{
					this.travelJumps.add(path);
				}
				
				// next becomes previous
				prev = next;
			}
		}
		
		logger.info("Ending waypoint calculation");
	}
	
	/**
	 * 
	 * @param start
	 * @param end
	 */
	private Vector getTravelPath(String start, String end)
	{
		logger.info("Starting travelpath calculation");
		// empty solarsystems
		SolarSystem sysStart, sysEnd;
		sysStart = null;
		sysEnd = null;
		Vector result = null;
		
		// loop through all solarsystems
		Enumeration num;
		num = this.solarSystems.elements();
		
		while(num.hasMoreElements())
		{
			SolarSystem current = (SolarSystem) num.nextElement();
			
			// is this solar system either our start or endpoint?
			if (current.name.equalsIgnoreCase(start))
			{
				sysStart = current;
			}
			else if (current.name.equalsIgnoreCase(end))
			{
				sysEnd = current;
			}
		}
		
		// both should not be null
		if (sysStart != null && sysEnd != null)
		{
			KwCalc calculator = new KwCalc();
			result = calculator.calcJumps(this.systemJumps, sysStart, sysEnd, this.avoidance, this.avoidSystems);
		}
		
		logger.info("Ending travelpath calculation");
		return result;
	}
	
	public void setRegions(Hashtable regions)
	{
		this.regions = regions;
	}
	
	public Hashtable getRegions()
	{
		return this.regions;
	}
}
